/*************************************************************************
* Copyright (c) 2004 Altera Corporation, San Jose, California, USA.      *
* All rights reserved. All use of this software and documentation is     *
* subject to the License Agreement located at the end of this file below.*
**************************************************************************
* Description:                                                           *
* The following is a simple hello world program running MicroC/OS-II.The * 
* purpose of the design is to be a very simple application that just     *
* demonstrates MicroC/OS-II running on NIOS II.The design doesn't account*
* for issues such as checking system call return codes. etc.             *
*                                                                        *
* Requirements:                                                          *
*   -Supported Example Hardware Platforms                                *
*     Standard                                                           *
*     Full Featured                                                      *
*     Low Cost                                                           *
*   -Supported Development Boards                                        *
*     Nios II Development Board, Stratix II Edition                      *
*     Nios Development Board, Stratix Professional Edition               *
*     Nios Development Board, Stratix Edition                            *
*     Nios Development Board, Cyclone Edition                            *
*   -System Library Settings                                             *
*     RTOS Type - MicroC/OS-II                                           *
*     Periodic System Timer                                              *
*   -Know Issues                                                         *
*     If this design is run on the ISS, terminal output will take several*
*     minutes per iteration.                                             *
**************************************************************************/

/*************************************************************************
 * Code used for initializing the pxiel buffer are modified from JunZhao  *
 * Appnotes in "tutorial of Displaying pixels on VGA monitor". It is     *
 * accessed at Feb 2, 2015 from
 * https://www.ualberta.ca/%7Edelliott/local/ece492/appnotes/2014w/G3_VGA/G3_VGA_PIXEL.qar
 **************************************************************************

/*************************************************************************
 * Modified by Ryan Corpuz, Jingjing Liang and Hang Peng                 *
 * Discription:                                                          *
 * This file shows the software layer of our design, The funcitionalities*
 * of this file includes calculating the pulse width of servos, controlling
 * movment of the servo and communicate between image processor and user *
 * interface                                                             *
 **************************************************************************/


#include <stdio.h>
#include <system.h>
#include <stdlib.h>
#include <time.h>
#include "includes.h"
#include "altera_up_avalon_video_pixel_buffer_dma.h"
#include "io.h"
#include "altera_up_avalon_character_lcd.h"

/* Definition of Colour */
#define   GREEN 0xF800

/* Definition of Image processor  Parameter */
#define H_MASK 0x00000007FFF
#define S_MASK 0x000007F8000
#define V_MASK 0x0007F800000

#define H_OFFSET 0
#define S_OFFSET 15
#define V_OFFSET 23

#define DEFAULT_H_RANGE 10
#define DEFAULT_S_RANGE 20
#define DEFAULT_V_RANGE 80

#define HSV_RANGE_REGISTER_OFFSET 24

#define H_REGISTER_MASK 0x000001ff
#define S_REGISTER_MASK 0x000000ff
#define V_REGISTER_MASK 0x000000ff

#define X_MASK 0x3ff
#define Y_MASK 0xFfC00
#define X_OFFSET 0
#define Y_OFFSET 10




/* Definition of Screen  */
#define   CENTER_X 160
#define   CENTER_Y 120
#define   MARGIN   5


/* Definition of Servo Moter Parameter*/
#define   MAX_PULSE_WIDTH 125000
#define   MIN_PULSE_WIDTH 75000
#define   PULSE_WIDTH_PER_DEGREE 425
#define   PIXEL_PER_DEGREE 26.5


/* Definition of Task Stacks */
#define   TASK_STACKSIZE       2048
OS_STK    task1_stk[TASK_STACKSIZE];
OS_STK    task2_stk[TASK_STACKSIZE];
OS_STK    task3_stk[TASK_STACKSIZE];
OS_STK    task4_stk[TASK_STACKSIZE];

/* Define Global Variable*/
alt_up_pixel_buffer_dma_dev * pixel_buf_dev;
alt_up_character_lcd_dev * character_lcd_dev;

float degree_x,degree_y;
int z = 20;
int flip =0;

/* Definition of Task Priorities */
#define TASK1_PRIORITY      1
#define TASK2_PRIORITY      2
#define TASK3_PRIORITY      3
#define TASK4_PRIORITY      4

/*
 * Servo control function for the HS-422
 * Input: The number of degrees of horizontal rotation needed
 * Output:
 * 		1  : If the servo angle is at max rotation (? ms pulse width)
 * 		-1 : If the servo angle is at min rotation (? ms pulse width)
 * 		0  : Servo doesn't reach limits
 */
int horizontal_servo(float number_of_degrees){
	// (-) left, (+) right
	// +- 950 to move 1 degree
	// Max is ~ 250000, min is ~ 60000
	int pulse_width = 0,err = 0;
	int old = IORD_32DIRECT(SERVO_0_BASE, 0);
	pulse_width = old + number_of_degrees*PULSE_WIDTH_PER_DEGREE;
    
	if (pulse_width > MAX_PULSE_WIDTH){
		pulse_width = MAX_PULSE_WIDTH;
		err = 1;
	}
	else if(pulse_width < MIN_PULSE_WIDTH){
		pulse_width = MIN_PULSE_WIDTH;
		err = -1;
	}

		old = IORD_32DIRECT(SERVO_0_BASE, 0);
		IOWR_32DIRECT(SERVO_0_BASE, 0, pulse_width);
		old = IORD_32DIRECT(SERVO_0_BASE, 0);
		return err;
}

/*
 * Servo control function for the HS-635HB
 * Input: The number of degrees of vertical rotation needed
 * Output:
 * 		1  : If the servo angle is at max rotation (? ms pulse width)
 * 		-1 : If the servo angle is at min rotation (? ms pulse width)
 * 		0  : Servo doesn't reach limits
 */
int vertical_servo(float number_of_degrees){
	// (-) up, (+) down
	// +- 950 to move 1 degree
	// Max is ~ 125000, min is ~ 25000
	int pulse_width = 0, err = 0, old = 0;
	old = IORD_32DIRECT(SERVO_1_BASE, 0);
	pulse_width = old - number_of_degrees*PULSE_WIDTH_PER_DEGREE;
	if (pulse_width > MAX_PULSE_WIDTH){
		pulse_width = MAX_PULSE_WIDTH;
		err = 1;
	}
	else if(pulse_width < MIN_PULSE_WIDTH){
		pulse_width = MIN_PULSE_WIDTH;
		err = -1;
	}
	old = IORD_32DIRECT(SERVO_1_BASE, 0);
	IOWR_32DIRECT(SERVO_1_BASE, 0, pulse_width);
	old = IORD_32DIRECT(SERVO_1_BASE, 0);
	return err;
}

/*
 * Initial_tracking function will be used to replace the value of the register which stores
 * threshold value in the image processor by the value read from the center of the frame.
 */
void inital_tracking(){
	int h_11, v_11, s_11, center_pixel,
        h_reg, s_reg, v_reg;

    // Read the center pixel from the image processor
    center_pixel = IORD_32DIRECT(IMAGE_PROCESSOR_0_P11_BASE,0);
    
    // Get H, S, V value of the center pxiel
	v_11 = (hsv_11&V_MASK)>>V_OFFSET;
	s_11 = (hsv_11&S_MASK)>>S_OFFSET;
	h_11 = (hsv_11&H_MASK)>>H_OFFSET;
    
    // Generate a 32bit number which contains the threshould range and threshold value of each colour fields
	h_reg = ((DEFAULT_H_RANGE<<HSV_RANGE_REGISTER_OFFSET) | (h_11&H_REGISTER_MASK));
	s_reg = ((DEFAULT_S_RANGE<<HSV_RANGE_REGISTER_OFFSET) | (s_11&S_REGISTER_MASK));
	v_reg = ((DEFAULT_V_RANGE<<HSV_RANGE_REGISTER_OFFSET) | (v_11&V_REGISTER_MASK));

    // Updating the value to the register.
	IOWR_32DIRECT(IMAGE_PROCESSOR_0_H_THRESH_BASE,0,h_reg);
	IOWR_32DIRECT(IMAGE_PROCESSOR_0_S_THRESH_BASE,0,s_reg);
	IOWR_32DIRECT(IMAGE_PROCESSOR_0_V_THRESH_BASE,0,v_reg);
    
    
    //Draw a green suqare at the center of the frame for indication
	alt_up_pixel_buffer_dma_draw_rectangle(pixel_buf_dev, CENTER_X-MARGIN, CENTER_Y-MARGIN, CENTER_X+MARGIN,CENTER_Y+MARGIN, GREEN, 0);

}

/*
 * This function is used to clear the LCD screen by erase the content of all the cursor postions
 */

void clear_lcd(alt_up_character_lcd_dev *lcd){
	int x = 0;
	int y = 0;
	alt_up_character_lcd_set_cursor_pos(lcd,0,0);
	for(y = 0; y < 2; y++){
		for(x = 0; x < 16; x++){
			alt_up_character_lcd_erase_pos(lcd,x,y);
		}
	}
	alt_up_character_lcd_set_cursor_pos(lcd,0,0);
}


/*
 * This function is used to write a string onto the LCD screen
 */
void write_lcd_msg(alt_up_character_lcd_dev *lcd, char* msg1, char* msg2){
	clear_lcd(lcd);
	alt_up_character_lcd_set_cursor_pos(lcd,0,0);
	alt_up_character_lcd_string(lcd, msg1);
	alt_up_character_lcd_set_cursor_pos(lcd,0,1);
	alt_up_character_lcd_string(lcd, msg2);
}

/*Clears the lcd*/
void servo_init(){
	IORD_32DIRECT(SERVO_1_BASE, 0);
	IOWR_32DIRECT(SERVO_1_BASE, 0, 100000);
	IORD_32DIRECT(SERVO_1_BASE, 0);
	IORD_32DIRECT(SERVO_0_BASE, 0);
	IOWR_32DIRECT(SERVO_0_BASE, 0, 75000);
	IORD_32DIRECT(SERVO_0_BASE, 0);

}

/*
 *This task is used to showing the message about the current states of the tracking system when
 *the configuration switch goes low.
 */
void task4(void* pdata){

	while(1){

		char* msg;
		char* msg2;
		if (!IORD_8DIRECT(CONFIG_SWITCH_BASE,0)){
			if (IORD_8DIRECT(H_LOCK_SWITCH_BASE,0)){
				msg = "X_LOCK is ON";
			}
			else{
				msg = "X_LOCK is OFF";
			}

			if (IORD_8DIRECT(V_LOCK_SWITCH_BASE,0)){
				msg2 = "Y_LOCK is ON";
			}
			else{
				msg2 = "Y_LOCK is OFF";
			}
			write_lcd_msg(character_lcd_dev,msg,msg2);
			OSTimeDlyHMSM(0, 0, 0, 250);
		}

	}
}

/*
 * This task is used to control the horizontal servo movement.
 */
void task2(void* pdata)
{
	while(1){
		if(!IORD_8DIRECT(H_LOCK_SWITCH_BASE,0) && !IORD_8DIRECT(CONFIG_SWITCH_BASE,0)){
			horizontal_servo(degree_x);

		}
		OSTimeDlyHMSM(0, 0, 0, z);
	}
}

/*
 * This task is used to control the vertical servo movement.
 */
void task3(void* pdata)
{
	while(1){
		if(!IORD_8DIRECT(V_LOCK_SWITCH_BASE,0)  && !IORD_8DIRECT(CONFIG_SWITCH_BASE,0)){
			vertical_servo(degree_y);
		}
		OSTimeDlyHMSM(0, 0, 0, z);
	}
}

/*
 * This task is used to retrieve and process the center of object in order to send movement instruction
 * to servo for tracking. It also handles two different states of the tracking system, setting the initial colour and tracking the colour.
 */
void task1(void* pdata)
{
	int r = 0, g = 0, b = 0;
	int xy = 0,h = 0, s = 0,v = 0,x = 0,y = 0, thr= 0;
	int counter = 0,row=0,row_index = 0;
	int temp1,temp2,temp3,temp4;
	int max,min;
	int xmin,xmax,ymin,ymax,relative_x,relative_y;
	character_lcd_dev = alt_up_character_lcd_open_dev("/dev/character_lcd_0");
	// open pixel buffer device
	pixel_buf_dev = alt_up_pixel_buffer_dma_open_dev("/dev/Pixel_Buffer_DMA");
		if (pixel_buf_dev == NULL)
			{
				// if pixel buffer cannot be found, type error message.
				printf("Error: could not open pixel buffer device \n");
			} else
			{
				printf("Opened pixel buffer device \n");
			}
				// Make sure the pixel buffer device has the primary buffer address
			if (pixel_buf_dev->buffer_start_address != 0)
			{
				alt_up_pixel_buffer_dma_change_back_buffer_address(pixel_buf_dev, 0);
				alt_up_pixel_buffer_dma_swap_buffers(pixel_buf_dev);
			}
			// swap buffer to the back buffer (start address of SDRAM for as frame buffer)
			alt_up_pixel_buffer_dma_swap_buffers(pixel_buf_dev);

            // Showing welcome message when the program is initialized properly
			alt_up_character_lcd_init(character_lcd_dev);
			char* welcome_string = "Hello, welcome!";
			alt_up_character_lcd_string(character_lcd_dev,welcome_string);
			alt_up_character_lcd_cursor_off(character_lcd_dev);
			OSTimeDlyHMSM(0, 0, 1, 0);


	while(1){

				if(IORD_8DIRECT(CONFIG_SWITCH_BASE,0)){
					flip = 1;
					servo_init();
                    // showing a brief instruction about setting colour
					char* msg = "Place object";
					char* msg2 ="in green square";
					write_lcd_msg(character_lcd_dev, msg, msg2);
                    // it will keep updating the threshold value in the register until the value
                    // of configuration switch goes low.
                    
					while(IORD_8DIRECT(CONFIG_SWITCH_BASE,0)){
						inital_tracking();
					}
                    // servo process will be suspended in setting state.
                    servo_init();

				}
        
                // If the state is tracking, then it will retrieve the center coordinate from register
                // in the image processor
				xy = IORD_32DIRECT(IMAGE_PROCESSOR_0_S1_BASE,0);
				x = (xy&X_MASK)>>X_OFFSET;
				y = (xy&Y_MASK)>>Y_OFFSET;


				relative_x = x - CENTER_X;
				relative_y = y - CENTER_Y;
        
                // IF the movement is not siginificant, then the servo can ignore it
				if ((relative_x < 10) && (relative_x > -10)){
					relative_x = 0;
				}
				if ((relative_y < 10) && (relative_y > -10)){
					relative_y = 0;
				}
        
				degree_x = relative_x/PIXEL_PER_DEGREE;
				degree_y = relative_y/PIXEL_PER_DEGREE;



		OSTimeDlyHMSM(0, 0, 0, 5);
		// 8 fps
	}
}


/* The main function creates two task and starts multi-tasking */
int main(void)
{
  servo_init();
  alt_up_character_lcd_init(character_lcd_dev);
  OSTaskCreateExt(task1,
                  NULL,
                  (void *)&task1_stk[TASK_STACKSIZE-1],
                  TASK1_PRIORITY,
                  TASK1_PRIORITY,
                  task1_stk,
                  TASK_STACKSIZE,
                  NULL,
                  0);

  OSTaskCreateExt(task2,
                    NULL,
                    (void *)&task2_stk[TASK_STACKSIZE-1],
                    TASK2_PRIORITY,
                    TASK2_PRIORITY,
                    task2_stk,
                    TASK_STACKSIZE,
                    NULL,
                    0);

  OSTaskCreateExt(task3,
                      NULL,
                      (void *)&task3_stk[TASK_STACKSIZE-1],
                      TASK3_PRIORITY,
                      TASK3_PRIORITY,
                      task3_stk,
                      TASK_STACKSIZE,
                      NULL,
                      0);
  OSTaskCreateExt(task4,
                        NULL,
                        (void *)&task4_stk[TASK_STACKSIZE-1],
                        TASK4_PRIORITY,
                        TASK4_PRIORITY,
                        task4_stk,
                        TASK_STACKSIZE,
                        NULL,
                        0);
              
               
  OSStart();
  return 0;
}



/******************************************************************************
*                                                                             *
* License Agreement                                                           *
*                                                                             *
* Copyright (c) 2004 Altera Corporation, San Jose, California, USA.           *
* All rights reserved.                                                        *
*                                                                             *
* Permission is hereby granted, free of charge, to any person obtaining a     *
* copy of this software and associated documentation files (the "Software"),  *
* to deal in the Software without restriction, including without limitation   *
* the rights to use, copy, modify, merge, publish, distribute, sublicense,    *
* and/or sell copies of the Software, and to permit persons to whom the       *
* Software is furnished to do so, subject to the following conditions:        *
*                                                                             *
* The above copyright notice and this permission notice shall be included in  *
* all copies or substantial portions of the Software.                         *
*                                                                             *
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  *
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    *
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE *
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      *
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     *
* FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         *
* DEALINGS IN THE SOFTWARE.                                                   *
*                                                                             *
* This agreement shall be governed in all respects by the laws of the State   *
* of California and by the laws of the United States of America.              *
* Altera does not recommend, suggest or require that this reference design    *
* file be used in conjunction or combination with any other product.          *
******************************************************************************/
